# -*- coding: utf-8 -*-
"""Configuration handling. Exports config class."""

import os
from pathlib import Path

import yaml

from .exceptions import ConfigNotFoundError
from .ui_messages import MISSING_CONF_ERROR_MESSAGE


class quareConfig:
    """Parses a configuration file and stores/retrieves its values."""

    DEFAULT_CONFIG_FILENAME = os.path.join(str(Path.home()), ".quare-conf.yaml")

    def __init__(self, config_path=None):
        self._find_config(config_path)
        self.raw_config = {}
        self._read_config()
        self._chats = []
        self._docs = []

    def _find_config(self, config_path):
        if config_path is not None and os.path.isfile(config_path):
            self.config_path = config_path
        else:
            self.config_path = self.DEFAULT_CONFIG_FILENAME

    def _read_config(self):
        if os.path.isfile(self.config_path):
            with open(self.config_path) as conf:
                self.raw_config = yaml.safe_load(conf.read())
            self.defaults = self.raw_config.get("defaults", {})
        else:
            raise ConfigNotFoundError(MISSING_CONF_ERROR_MESSAGE)

    @property
    def default_docs(self):
        if not self._docs:
            self._docs = self.defaults.get("documents", [])
        return self._docs

    @property
    def default_chats(self):
        if not self._chats:
            self._chats = self.defaults.get("chats", [])
        return self._chats



# -*- coding: utf-8 -*-
"""Quip objects. Contains filtering logic, but no API calls."""

import time
import click
from dataclasses import dataclass
from typing import Any, Dict, List

import quare.quip as quip


@dataclass
class Meta:
    """Quip stores metadata in a `meta` property. This is the base class for common properties."""

    id: str
    created_usec: int
    updated_usec: int


@dataclass
class FolderMeta(Meta):
    title: str
    color: str = "manila"
    parent_id: str = None
    creator_id: str = None


@dataclass
class ThreadMeta(Meta):
    link: str
    title: str
    author_id: str
    thread_class: Any = None
    sharing: str = None
    type: Any = None
    document_id: str = None


@dataclass
class QuipFolder:
    member_ids: List[str]
    children: List[Dict[str, str]]
    folder: Dict

    def __post_init__(self):
        """Unpack the folder dictionary into the meta field."""
        self.meta = FolderMeta(**self.folder)

    @property
    def child_threads(self):
        return self._filter_children("thread_id")

    @property
    def child_folders(self):
        return self._filter_children("folder_id")

    def _filter_children(self, key):
        return [child[key] for child in self.children if key in child]


@dataclass
class QuipMessage:
    """Represents a quip message."""

    id: str
    created_usec: int
    updated_usec: int
    text: str
    author_name: str
    author_id: str
    visible: str
    annotation: Dict = None
    parts: List[List[str]] = None
    like_user_ids: List[str] = None
    mention_user_ids: List[str] = None
    files: List[str] = None

    @classmethod
    def from_dict(cls, dict_val):
        return cls(**{k: v for k, v in dict_val.items() if k in cls.__annotations__})

    @property
    def created(self):
        return self._parse_time(self.created_usec)

    @property
    def updated(self):
        return self._parse_time(self.updated_usec)

    def _parse_time(self, usecs):
        return time.asctime(time.localtime(usecs / 1e6))

    def __hash__(self):
        """
        On the remote chance that we have two messages with the same created_usec,
        sum the characters in that message's ID and add it to the created_usec.
        Almost certainly a case of YAGNI.
        """
        return self.created_usec + sum([ord(c) for c in self.id])

    def print(self, style=True, thread=None):
        thread = f"| ({thread}) " if thread else ""
        msg_header = f"[{self.created} {thread}| @{self.author_name}]"
        msg_header = click.style(msg_header, bold=True) if style else msg_header
        click.echo(f"{msg_header} {self.text}")


@dataclass
class QuipThread:
    """A Quip Thread. Either a chat or document.

    Args:
        user_ids (List[str]): User IDs with access to this thread.
        shared_folder_ids (List[str]): List of QuipFolder IDs containing this thread.
        expanded_user_ids (List[str]): User IDs with access to this thread.
        invited_user_emails (List[str]): Emails of users with access to this thread.
        html (str): The contents as HTML, if this thread is a document.
        thread (Dict): Dictionary containing metadata
        meta (ThreadMeta): ThreadMeta instance containing metadata
        token (str): API access token.
    """

    user_ids: List[str]
    thread: Dict
    shared_folder_ids: List[str] = None
    expanded_user_ids: List[str] = None
    invited_user_emails: List[str] = None
    html: str = None

    def __post_init__(self):
        """Unpack the folder dictionary into the meta field."""
        self.meta = ThreadMeta(**self.thread)
        self.id = self.meta.id

        # Instance variables
        self.token = None

    @staticmethod
    def get_thread(token, room):
        """Instantiate a thread by calling the QuipAPI."""
        client = quip.QuipClient(token)
        thread_json = client.get_thread(room)
        return QuipThread(**thread_json)

    def get_messages(
        self, token=None, start_usec=1000, last_n=None, return_raw_json=False
    ):
        """Returns the most recent messages for this thread.

        Args:
            start_usec: Return all messages since this timestamp.
            last_n: An integer indicating the number of messages you want returned.
                This is distinct from the `count` param accepted by the Quip API,
                in that it is the total number of messages to return, not the
                total to return for a given request.

        Returns:
            A list of message dicts, sorted by their created date in ascending order
        """
        self.messages = []
        client = quip.QuipClient(token)
        COUNT = 200
        if last_n and last_n < COUNT:
            COUNT = last_n
        last_usec = None

        while True:
            # I think the non list comprehension version is equally unreadable
            message_dicts = [
                msg
                for msg in client.get_messages(
                    self.meta.id, max_created_usec=last_usec, count=last_n
                )
                if msg["created_usec"] > start_usec
            ]

            last_count = len(message_dicts)
            last_usec = message_dicts[-1]["created_usec"]
            self.messages.extend(message_dicts)
            if last_count in [last_n, 1] or (last_n and len(self.messages) >= last_n):
                break
        return self.messages

    def send_message(
        self,
        token=None,
        frame=None,
        content=None,
        silent=False,
        return_raw_json=False,
        parts=None,
    ) -> QuipMessage:
        """Send a message to this thread.

        Args:
            frame: Controls how Quip displays the message.
            body: An integer indicating the number of messages you want returned.
                This is distinct from the `count` param accepted by the Quip API,
                in that it is the total number of messages to return, not the
                total to return for a given request.

        Returns:
            A QuipMessage instance
        """
        client = quip.QuipClient(token)
        # Using the vendored library, note that it passes keywords to the api via **kwargs
        msg_dict = client.new_message(
            self.id, content=content, silent=silent, frame=frame, parts=parts
        )
        return QuipMessage.from_dict(msg_dict)



# -*- coding: utf-8 -*-

"""Console messages to the user for quare."""

import click

_ERROR_COLOR = "bright_red"
_ERROR_MSG_COLOR = "magenta"


def _error(msg):
    """Applies default error header style."""
    return click.style(msg, fg=_ERROR_COLOR, bold=True)


# console messages printed via ClickException or Quip*
AUTH_ERROR_MESSAGE = f"""
{_error('Quip says this is an invalid access token.')} To generate a new one to add via `quare auth', visit: https://quip.com/dev/token
"""
TOKEN_ERROR_MESSAGE = f"""
{_error('No default token is set.')} Use `quare auth' or set a QUIP_TOKEN  environment variable containing your authentication token.
"""
MISSING_CONF_ERROR_MESSAGE = f"""{_error('Config is not a file.')} Check your path."""



# -*- coding: utf-8 -*-

"""Console script for quare."""

import json
import html

import click

import quare.utils as utils

from .quare import get_user_info, get_user_token, set_user_token
from .quip import QuipClient, QuipError
from .quip_classes import QuipMessage, QuipThread
from .quip_websocket import stream_updates
from .ui_messages import AUTH_ERROR_MESSAGE


@click.group("main", help="")
@click.version_option()
def main():
    """Main quare CLI entry point.

    This runs as the first step in processing any CLI command.
    """


@click.command()
@click.option("-a", "--alias", default="Default", show_default=True)
@click.option("--json", "json_", is_flag=True)
def whoami(alias, json_):
    """Print information about the logged in users."""
    try:
        alias, token = get_user_token(alias)
        user, user_table = get_user_info(token, alias=alias)
        if json_:
            click.echo(json.dumps(user))
        else:
            click.echo(user_table)
    except QuipError as ex:
        if ex.code == 401 and "Invalid access_token" in str(ex):
            raise click.ClickException(AUTH_ERROR_MESSAGE)
        raise  # reraise


@click.command()
@click.option("-a", "--alias", default="Default")
@click.option("-t", "--token", prompt=True)
def auth(alias, token):
    """Store an API authentication token."""
    set_user_token(alias, token.strip())
    click.secho("Token stored!", fg="bright_green")


main.add_command(whoami)
main.add_command(auth)


@main.group("doc")
def document():
    """Commands for interacting with documents."""


@document.command(name="append")
@click.option(
    "-a", "--alias", default="Default", help="Quip `auth' alias", show_default=True
)
@click.option("-c", "--content", help="Markdown to append to document.")
@click.option(
    "-f",
    "--file",
    "fd",
    type=click.File(),
    help="Markdown to append to document. Will ignore --content if provided.",
)
@click.option("-i", "--id", "doc_id", required=True)
@click.option("--json", "json_", is_flag=True)
def doc_append(alias, content, fd, doc_id, json_):
    """Append content to an existing Quip document."""
    _, token = get_user_token(alias)
    content = fd.read() if fd else content
    if not content:
        raise click.UsageError('You must pass either "--content" or "--file".')
    client = QuipClient(token)
    client.edit_document(doc_id, f"\n{content}\n", format="markdown")


@main.group("msg")
def message():
    """Commands for interacting with chats/comment threads."""


@message.command(name="get")
@click.option("-r", "--room", required=True)
@click.option("-s", "--since", type=str, default="1999")
@click.option("-a", "--alias", default="Default", show_default=True)
@click.option("--json", "json_", is_flag=True)
@click.option(
    "-l", "--last", default=200, show_default=True, type=int, help="Last n messages"
)
@click.option(
    "--decending/--ascending",
    "is_desc",
    default=False,
    help="Created date/time sort order",
)
def msg_get(room, last, since, alias, json_, is_desc):
    """Retrieve messages from Quip."""
    token, thread = get_thread(alias, room)
    start_usec = utils.get_usec(since)
    messages = thread.get_messages(token, last_n=last, start_usec=start_usec)
    if not is_desc:
        messages.reverse()
    print_msgs(messages, json_)


@message.command(name="stream")
@click.option("-a", "--alias", default="Default", show_default=True)
def msg_stream(alias):
    """Stream messages via Quip's websocket. Close using CTRL-C"""
    _, token = get_user_token(alias)
    stream_updates(token)


@message.command(name="send")
@click.option("-a", "--alias", default="Default", show_default=True)
@click.option(
    "-c", "--content", help="The body of the message to send. Use '-' for stdin"
)
@click.option("-r", "--room", required=True)
@click.option(
    "--frame",
    type=click.Choice(["bubble", "card", "line"]),
    help="Adjust the display of the message",
)
@click.option("--silent", is_flag=True)
@click.option(
    "-m", "--monospace", help="Format the message as monospace code.", is_flag=True
)
@click.option("--file", "file_", type=click.File("r"))
def msg_send(alias, content, room, frame, silent, monospace, file_):
    """Send messages to a quip chat/document."""
    token, thread = get_thread(alias, room)
    content = click.get_text_stream("stdin").read() if content == "-" else content
    content = file_.read() if file_ else content
    parts = format_msg_content(content, monospace=monospace)
    msg = thread.send_message(
        token, content=content, frame=frame, silent=silent, parts=parts
    )
    if msg:
        messages = thread.get_messages(token, last_n=5)
        messages.reverse()
        print_msgs(messages)


def format_msg_content(content, monospace=False):
    if monospace:
        content = (
            "<pre class='prettyprint'>"
            + html.escape(content, quote=True).replace("\n", "<br>").replace(" ", " ")
            + "</pre>"
        )
        return json.dumps([["monospace", content]])
    else:
        return None


def get_thread(alias, room):
    """Send messages to a quip chat/document."""
    _, token = get_user_token(alias)
    thread = QuipThread.get_thread(token, room)
    return token, thread


def print_msgs(messages, json_=False):
    """Print and format messages."""
    if json_:
        click.echo(json.dumps(messages))
        return
    else:
        for message in messages:
            utils.print_quip_message(QuipMessage.from_dict(message))



"""Custom exceptions for quare and QuipClient"""


class QuipTokenNotFoundError(Exception):
    """
    Thrown if a Quip token is not in the keyring, and the QUIP_TOKEN environment variable is not set."""


class ConfigNotFoundError(Exception):
    """Thrown if a quare config file does not exist."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quare']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.0,<8.0',
 'colorama>=0.4.1,<0.5.0',
 'dateparser>=0.7.1,<0.8.0',
 'keyring>=18.0,<19.0',
 'terminaltables>=3.1,<4.0',
 'websocket-client>=0.56.0,<0.57.0']

entry_points = \
{'console_scripts': ['quare = quare.cli:main']}

setup_kwargs = {
    'name': 'quare',
    'version': '0.2.4',
    'description': 'quare is a CLI client for Quip.',
    'long_description': '# quare\n\n[![pipeline status](https://gitlab.com/jstvz/quare/badges/master/pipeline.svg)](https://gitlab.com/jstvz/quare/commits/master)\n[![coverage report](https://gitlab.com/jstvz/quare/badges/master/coverage.svg)](https://gitlab.com/jstvz/quare/commits/master)\n[![PyPI - License](https://img.shields.io/pypi/l/quare.svg)](https://www.gnu.org/licenses/lgpl-3.0.en.html)\n![PyPI](https://img.shields.io/pypi/v/quare.svg)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/quare.svg)\n![PyPI - Status](https://img.shields.io/pypi/status/quare.svg)\n\nInteract with Quip from the command line.\n\n![quare streaming messages](https://github.com/jstvz/quare/blob/master/docs/assets/quare.png?raw=true)\n\n## Introduction ##\nquare allows interaction with [Quip](https://quip.com) via the command line. While `quare` is in alpha, there are some features you may find useful:\n- Pipe the output of a command to a chat or document and format it as monospace.\n- Archive messages by piping them into a local file\n- Securely store authentication tokens for multiple Quip instances.\n\nThis tool is in its early stages of development, and is subject to change (or abandonment) at any time. Use at your own risk.\n\n## Installation ##\n\n```console\n$ pipx install quare\n```\n\n## Usage ##\n\n### Authentication ###\nStore a Quip API token (See: https://quip.com/dev/token):\n\n```console\n$ quare auth\nToken: long_token_string\nToken stored.\n```\nIf you have multiple Quip instances (like multiple Slack Workspaces), you can specify an alias for them. You can also pass your token directly to `auth`:\n\n```console\n$ quare msg auth --alias test_server --token \'t1DJBQWBXHCYgh1=|2983928392|nYtRFIhV7nl4...\'\n```\n\n#### Whoami ####\n\nTo see information about the logged in user:\n```console\n$ quare msg whoami\n┌Default───────┬───────────────┐\n│ Name         │ Tests Testeri │\n├──────────────┼───────────────┤\n│ Email(s)     │ t@testz.dev   │\n├──────────────┼───────────────┤\n│ Quip User ID │ mRLA6Zdn3PO   │\n└──────────────┴───────────────┘\n```\n\n### Sending messages ###\nThe destination may be a document or chat:\n\n```console\n$ quare msg send --room room_id --content \'Hello everyone!\'\n```\n\n#### Pipe content from `stdin` ####\n\nMessage content can be piped from `stdin`:\n```console\n$ uname -a | quare msg send --room room_id --content \'-\'\n```\n\nWhile Quip allows formatting messages using some markdown markup, it doesn\'t recognize markdown code blocks ("\\`\\`\\`"). To define a code block, use the `--monospace` option:\n\n```console\n$ dmesg | tail -n 5 | quare msg send --room room_id --content \'-\' --monospace\n```\n\n### Receiving messages ###\n\n#### Stream to stdout ####\nTo stream every message appearing in the Updates tab:\n\n```console\n$ quare msg stream\nStreaming updates... press Ctrl-C to exit.\n[Sun Jun 30 17:23:09 2019 | (Test Log) | @Tests Testeri] ok ok\n```\n\n#### Dump the content of a chat room ####\nTo get the last 5 messages in a chat room or document:\n```console\n$ quare msg get --room room_id --last 5\n[Sat Jun 29 03:19:09 2019 | @Tester Testeri] This is a monologue!\n[Sat Jun 29 16:00:12 2019 | @Tester Testeri] ok\n[Sat Jun 29 16:34:51 2019 | @Tester Testeri] I\'m done!\n[Sun Jun 30 17:30:14 2019 | @Tester Testeri] 🌮\n[Sun Jun 30 17:30:27 2019 | @Tester Testeri] 🥃\n```\n\nTo get the last 2 messages as JSON:\n```console\n$ quare msg get  -r "IcTAAAtVxXb" --last 2 --json\n[{"author_id": "mRLA6Zdn3PO", "visible": true, "id": "IcderpEe8wG", "created_usec": 1561849212672040, "updated_usec": 1561849212696571, "text": "ok", "author_name": "Tester Testeri"}, {"author_id": "mRLA6Zdn3PO", "visible": true, "id": "IcNodg7n2Tx", "created_usec": 1561851291612434, "updated_usec": 1561851291620308, "text": "chat", "author_name": "Tester Testeri"}]\n```\n\nTo dump the last 200 messages in a chat room into a text file:\n```console\n$ quare msg get --room room_id --last 200 > interesting_conversation.log\n```\n\nTo get all messages since a datetime:\n```console\n$ quare msg get --room room_id --since 2019-01-01T00:32:00Z > greppable_archive.log\n```\n\nThe `--since` option recognizes any date recognized by [dateparser](https://dateparser.readthedocs.io/en/latest/):\n```console\n$ quare msg get --room room_id --since \'Monday\' > this_week.log\n$ quare msg get --room room_id --since \'2 months ago\' > this_quarter.log\n```\n\n### Editing documents ###\n\nTo append a markdown file to an existing document\n```console\n$ quare doc append --id xxxDoc_IDxxx --file /tmp/foo.md\n$ cat /tmp/foo.md | quare doc append --id xxxDoc_IDxxx --file -\n```\n\nTo append a markdown-format string to an existing document\n```console\n$ quare doc append --id xxxDoc_IDxxx --content \'## Headline\\n\\n\'\n```\n\n## Development ##\n\nThis work is licensed under the terms of the [LGPL-3.0](https://www.gnu.org/licenses/lgpl-3.0.en.html).\n### Contributions ###\nSee CONTRIBUTING.rst\n',
    'author': 'James Estevez',
    'author_email': 'j@jstvz.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jstvz/quare',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Contains business logic for interactions with the Quip API."""

import os

import keyring
import terminaltables

from quare import quip, ui_messages
from quare.exceptions import QuipTokenNotFoundError
from quare.quip_classes import QuipFolder

SVC_NAME = "quare"
TOKEN_ENV_VAR = "QUIP_TOKEN"  # nosec


def get_user_token(alias="Default"):
    """Retreive an api token from the keyring or environment variable.

    Keyword args:
    alias -- the alias under which the token is stored (default "Default")
    """
    token = keyring.get_password(SVC_NAME, alias)
    if type(token) is str:
        return alias, token
    elif TOKEN_ENV_VAR in os.environ and os.environ[TOKEN_ENV_VAR]:
        return alias, os.environ[TOKEN_ENV_VAR]
    else:
        raise QuipTokenNotFoundError(ui_messages.TOKEN_ERROR_MESSAGE)


def set_user_token(alias, token):
    """Store an API token and associated metadata."""
    keyring.set_password(SVC_NAME, alias, token)


def get_user_info(token, alias="User"):
    """Call the Quip API and get information about the authenticated user.

    Keyword args:
    alias -- the alias under which the token is stored (default "Default")
    """
    client = quip.QuipClient(token)
    user = client.get_authenticated_user()
    table = terminaltables.SingleTable(_format_user_matrix(user), alias)
    table.inner_row_border = True
    return user, table.table


def send_messages(token, alias, room):
    """Call the Quip API and send a message."""
    client = quip.QuipClient(token)
    messages = client.send_messages(room)
    return messages


def _format_user_matrix(user):
    """Given a Quip user dict, return a matrix that terminal tables will accept."""
    return [
        ["Name", user["name"]],
        ["Email(s)", ", ".join(user["emails"]) if "emails" in user else ""],
        ["Quip User ID", user["id"]],
    ]


def _get_user_favorites(user, token):
    """Given a user dict, recursively retrieve the user's starred items."""
    starred_id = user["starred_folder_id"]
    client = quip.QuipClient(token)
    folder = client.get_folder(starred_id)
    print(folder)
    folder = QuipFolder(**folder)
    return _get_children(folder, token)


def _get_children(folder, token):
    """Get all child threads for this folder."""
    client = quip.QuipClient(token)
    threads = client.get_threads(folder.child_threads)
    return threads
